import json
import random

import contextily as cx
import geojson
from matplotlib.lines import lineStyles
import matplotlib.pyplot as plt
from numpy import poly
import shapely
import trajgenpy.Logging
from trajgenpy import Utils
from trajgenpy.Geometries import (
    GeoMultiPolygon,
    GeoMultiTrajectory,
    GeoPolygon,
    GeoTrajectory,
    decompose_polygon,
    generate_sweep_pattern,
    get_sweep_offset,
)
from trajgenpy.Query import query_features


log = trajgenpy.Logging.get_logger()


def get_line_segments(polygon: GeoPolygon):
    polygon_list = decompose_polygon(polygon.get_geometry())

    log.info(f"Number of polygons: {len(polygon_list)}")

    offset = get_sweep_offset(0.1, 15, 90)
    result = []
    for decomposed_poly in polygon_list:
        sweeps_connected = generate_sweep_pattern(
            decomposed_poly, offset, connect_sweeps=False
        )
        result.extend(sweeps_connected)
    return result


def export_to_geojson(tasks: GeoMultiTrajectory, polygon: GeoPolygon):
    log.info(f"Number of sweeps: {len(tasks.get_geometry().geoms)}")
    geojson_collection = geojson.FeatureCollection(
        [
            polygon.to_geojson(id="boundary"),
            multi_traj.to_geojson(id="tasks"),
        ]
    )
    with open("environment.geojson", "w") as f:
        geojson.dump(geojson_collection, f)


if __name__ == "__main__":
    polygon_file = "examples/DemaScenarios/FlatTerrainNature.geojson"
    # polygon_file = "examples/DemaScenarios/HillyTerrainNature.geojson"
    # polygon_file = "examples/DemaScenarios/Urban.geojson"
    # polygon_file = "examples/DemaScenarios/Water.geojson"

    # Load the GeoJSON data
    with open(polygon_file, "r") as f:
        data = json.load(f)

    # Extract the polygon coordinates
    coordinates = data["features"][0]["geometry"]["coordinates"][0]

    # Create the shapely Polygon object
    polygon = GeoPolygon(shapely.Polygon(coordinates)).set_crs("EPSG:2197")

    tags = {
        "natural": ["water", "wetland"],
    }

    wetland = query_features(
        GeoPolygon(shapely.Polygon(coordinates)),
        {
            "natural": ["water", "wetland"],
        },
    )
    for f in wetland.values():
        GeoMultiPolygon(f).set_crs("EPSG:2197").plot()

    roads = query_features(
        GeoPolygon(shapely.Polygon(coordinates)),
        {
            "highway": ["service", "track"],
        },
    )
    for road in roads.values():
        GeoMultiTrajectory(road).set_crs("EPSG:2197").plot(color="red")

    tasks = get_line_segments(polygon)

    multi_traj = GeoMultiTrajectory(tasks, "EPSG:2197")
    multi_traj.set_crs("EPSG:2197")
    polygon.set_crs("EPSG:2197")

    # remove duplicate lines
    trajectories = multi_traj.get_geometry().geoms
    duplicates = []
    for i in range(len(multi_traj.get_geometry().geoms)):
        for j in range(i + 1, len(multi_traj.get_geometry().geoms)):
            if (
                multi_traj.get_geometry().geoms[i].coords[0]
                == multi_traj.get_geometry().geoms[j].coords[-1]
            ):
                duplicates.append(multi_traj.get_geometry().geoms[i])
                print("found duplicate")
    trajectories = [traj for traj in trajectories if traj not in duplicates]
    multi_traj = GeoMultiTrajectory(trajectories, "EPSG:2197")
    # multi_traj.plot(color="red", linestyle="dashed")

    polygon.plot(facecolor="none", edgecolor="black", linewidth=2)
    polygon.buffer(50).plot(
        facecolor="none",
        edgecolor="orange",
        linewidth=2,
        linestyle="dashed",
    )

    polygon.buffer(100).plot(
        facecolor="none",
        edgecolor="red",
        linewidth=2,
        linestyle="dashed",
    )

    Utils.plot_basemap(provider=cx.providers.CartoDB.Positron, crs="EPSG:2197")
    export_to_geojson(multi_traj, polygon)

    # No axis on the plot
    # plt.axis("equal")
    plt.axis("off")
    plt.show()



import geojson
import matplotlib.pyplot as plt
import shapely
from trajgenpy import Utils
from trajgenpy.Geometries import (
    GeoMultiPolygon,
    GeoMultiTrajectory,
    GeoPolygon,
    decompose_polygon,
    generate_sweep_pattern,
    get_sweep_offset,
)
from trajgenpy.Query import query_features


if __name__ == "__main__":
    # Download the water features data within the bounding box
    tags = {
        # "natural": ["water", "wetland"],
        "natural": ["coastline"],
        # "landuse": ["farmland"],
        "highway": [
            "primary",
            "secondary",
            "secondary_link",
            "tertiary",
            "unclassified",
            "road",
            "service",
            "path",
            "track",
        ],
        "building": True,
        "height": True,
    }

    # Amagerværket: shapely.Polygon([(12.620400,55.687962),(12.632788,55.691589),(12.637446,55.687689),(12.624924,55.683489)])
    # Davinde: shapely.Polygon([(10.490913, 55.315346), (10.576744, 55.315346), (10.576744, 55.337417), (10.490913, 55.337417)])
    polygon = GeoPolygon(
        shapely.Polygon(
            [
                (12.620400, 55.687962),
                (12.632788, 55.691989),
                (12.637446, 55.687689),
                (12.624924, 55.683489),
            ]
        ),
        crs="WGS84",
    )  # Download the water features data within the bounding box

    features = query_features(polygon, tags)
    polygon.set_crs("EPSG:2197")

    # polygon.plot(facecolor="none", edgecolor="black", linewidth=2)

    # Plot highways
    roads = GeoMultiTrajectory(features["highway"], crs="WGS84").set_crs("EPSG:2197")
    # roads.plot()

    # Plot Buildings
    obstacles = GeoMultiPolygon(features["building"], crs="WGS84").set_crs("EPSG:2197")
    obstacles.buffer(5)
    # obstacles.plot(color="red")

    # Plot natural features
    coastline = GeoPolygon(features["natural"], crs="WGS84").set_crs("EPSG:2197")
    coastline.plot(color="green", facecolor="none")

    extra_search = GeoPolygon(
        shapely.Polygon(
            [
                [12.63313389260452, 55.688680108812235],
                [12.631452519091368, 55.68805370657546],
                [12.632195079795991, 55.68729833042721],
                [12.633739947469763, 55.68788531188278],
                [12.633099751882298, 55.68842992615794],
                [12.63313389260452, 55.688680108812235],
            ]
        ),
        crs="WGS84",
    ).set_crs("EPSG:2197")
    extra_search.plot(color="green")

    polygon_list = decompose_polygon(
        coastline.get_geometry(), obstacles=extra_search.get_geometry()
    )

    # GeoMultiPolygon(
    #     shapely.MultiPolygon(polygon_list), crs="EPSG:2197"
    # ).plot(facecolor="none", edgecolor="blue", linewidth=2)

    # Export the polygon and the buildings as a GeoJSON file
    geojson_collection = geojson.FeatureCollection(
        [
            polygon.to_geojson(id="boundary"),
            obstacles.to_geojson(id="obstacles"),
        ]
    )

    offset = get_sweep_offset(0.15, 10, 90)
    result = []
    for decomposed_poly in polygon_list:
        sweeps_connected = generate_sweep_pattern(
            decomposed_poly, offset, clockwise=True, connect_sweeps=True
        )

        # Plotting
        result.extend(sweeps_connected)
    multi_traj = GeoMultiTrajectory(result, "EPSG:3857")
    multi_traj.plot(color="red")

    with open("environment.geojson", "w") as f:
        geojson.dump(geojson_collection, f)

    # Plot on a map
    Utils.plot_basemap(crs="EPSG:2197")

    # No axis on the plot
    plt.axis("equal")
    plt.axis("off")
    plt.show()



import math

import pytest
from shapely.geometry import LineString, Point, Polygon
from trajgenpy import Geometries, Logging

log = Logging.get_logger()

# Import your class from the 'trajectory' module


# TESTs
# Plot trajectories:
# import matplotlib.pyplot as plt
#     # plot test using matplotlib
#     # Extract the x and y coordinates for plotting
#     fig, ax = plt.subplots()

#     for multilinestring in list(test.geoms):
#         x, y = multilinestring.xy
#         ax.plot(x, y)
#     plt.show()

# Plot polygons:
# import matplotlib.pyplot as plt
#     # plot the shapely multi_polygon
#     fig, ax = plt.subplots()
#     for polygon in multi_polygon:
#         x, y = polygon.exterior.xy
#         ax.plot(x, y)

#     x, y = geo_poly.get_geometry().exterior.xy
#     ax.plot(x, y)
#     plt.show()


# Test initialization and conversion for Trajectory class
def test_trajectory():
    linestring = LineString(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
        ]
    )
    trajectory = Geometries.GeoTrajectory(linestring)

    # Check initial CRS
    assert trajectory.crs == "WGS84"

    # Convert to a different CRS
    trajectory.set_crs("EPSG:3857")

    # Check the new CRS
    assert trajectory.crs == "EPSG:3857"

    # Check converted geometry
    converted_coords = list(trajectory.get_geometry().coords)
    assert pytest.approx(converted_coords) == [
        (1404896.5016074297, 7496546.845393788),
        (1406275.5274593767, 7497263.139176297),
        (1406794.053647492, 7496492.933496759),
        (1405400.1109837785, 7495663.567131141),
    ]


# Test initialization and conversion for PointData class
def test_point_data():
    point = Point(12.624924, 55.683489)
    point_data = Geometries.GeoPoint(point)

    # Check initial CRS
    assert point_data.crs == "WGS84"

    # Convert to a different CRS
    point_data.set_crs("EPSG:3857")

    # Check the new CRS
    assert point_data.crs == "EPSG:3857"

    # Check converted geometry
    assert point_data.get_geometry() == Point(1405400.1109837785, 7495663.567131141)


# Test initialization and conversion for PolygonData class
def test_polygon_data():
    polygon = Polygon(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
        ]
    )
    polygon_data = Geometries.GeoPolygon(polygon)

    # Check initial CRS
    assert polygon_data.crs == "WGS84"

    # Convert to a different CRS
    polygon_data.set_crs("EPSG:3857")

    # Check the new CRS
    assert polygon_data.crs == "EPSG:3857"

    # Check converted geometry
    converted_coords = list(polygon_data.get_geometry().exterior.coords)
    assert pytest.approx(converted_coords) == [
        (1404896.5016074297, 7496546.845393788),
        (1406275.5274593767, 7497263.139176297),
        (1406794.053647492, 7496492.933496759),
        (1405400.1109837785, 7495663.567131141),
        (1404896.5016074297, 7496546.845393788),
    ]


def test_valid_inputs():
    assert math.isclose(Geometries.get_sweep_offset(0.1, 30, 90), 54, abs_tol=1e-3)
    assert math.isclose(Geometries.get_sweep_offset(0.5, 40, 120), 69.282, abs_tol=1e-3)


def test_invalid_overlap():
    with pytest.raises(
        ValueError, match="Overlap percentage has to be a float between 0 and 1!"
    ):
        Geometries.get_sweep_offset(1.2, 30, 90)


def test_negative_overlap():
    with pytest.raises(
        ValueError, match="Overlap percentage has to be a float between 0 and 1!"
    ):
        Geometries.get_sweep_offset(-0.1, 30, 90)


def test_decompose():
    poly = Polygon(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
            (12.628446, 55.686489),
            (12.625924, 55.688489),
            (12.630924, 55.689489),
        ]
    )

    geo_poly = Geometries.GeoPolygon(poly)
    geo_poly.set_crs("EPSG:3857")

    polygon_list = Geometries.decompose_polygon(geo_poly.get_geometry(), obstacles=None)

    # Define the coordinates for the hole (inner polygon)

    hole = Polygon(
        [
            (12.629, 55.688),
            (12.631, 55.689),
            (12.632, 55.687),
        ]
    )

    hole = Geometries.GeoPolygon(hole)
    hole.set_crs("EPSG:3857")

    polygon_list = Geometries.decompose_polygon(
        geo_poly.get_geometry(), obstacles=hole.get_geometry()
    )

    # Assert that the sum of areas of the decomposed polygons is equal to the area of the original polygon
    total_area = geo_poly.get_geometry().area - hole.get_geometry().area
    assert pytest.approx(sum([poly.area for poly in polygon_list])) == total_area
    assert len(polygon_list) > 0


def test_sweep_gen():
    offset = Geometries.get_sweep_offset(0.1, 30, 90)
    poly = Polygon(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
        ]
    )
    geo_poly = Geometries.GeoPolygon(poly)
    geo_poly.set_crs("EPSG:3857")

    test = Geometries.generate_sweep_pattern(
        geo_poly.get_geometry(), offset, clockwise=False, connect_sweeps=True
    )
    assert len(test) == 1


def test_sweep_gen_with_obstacle():
    poly = Polygon(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
            (12.628446, 55.686489),
            (12.625924, 55.688489),
            (12.630924, 55.689489),
        ]
    )

    geo_poly = Geometries.GeoPolygon(poly)
    geo_poly.set_crs("EPSG:3857")

    polygon_list = Geometries.decompose_polygon(geo_poly.get_geometry(), obstacles=None)

    # Define the coordinates for the hole (inner polygon)

    hole = Polygon(
        [
            (12.629, 55.688),
            (12.631, 55.689),
            (12.632, 55.687),
        ]
    )

    hole = Geometries.GeoPolygon(hole)
    hole.set_crs("EPSG:3857")

    polygon_list = Geometries.decompose_polygon(
        geo_poly.get_geometry(), obstacles=hole.get_geometry()
    )

    offset = Geometries.get_sweep_offset(0.1, 30, 90)
    for decomposed_poly in polygon_list:
        sweeps_connected = Geometries.generate_sweep_pattern(
            decomposed_poly, offset, clockwise=True, connect_sweeps=True
        )
        assert len(sweeps_connected) == 1

        sweeps_disconnected = Geometries.generate_sweep_pattern(
            decomposed_poly, offset, clockwise=True, connect_sweeps=False
        )
        assert len(sweeps_disconnected) != 1


def test_shapely_polygon_to_cgal():
    poly = Polygon(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
        ]
    )
    cgal_poly = Geometries.shapely_polygon_to_cgal(poly)
    assert cgal_poly is not None


def test_obstacle_polygon_overlaps_boundary():
    poly = Polygon(
        [
            (12.620400, 55.687962),
            (12.632788, 55.691589),
            (12.637446, 55.687689),
            (12.624924, 55.683489),
            (12.628446, 55.686489),
            (12.625924, 55.688489),
            (12.630924, 55.689489),
        ]
    )

    geo_poly = Geometries.GeoPolygon(poly)
    geo_poly.set_crs("EPSG:3857")

    polygon_list = Geometries.decompose_polygon(geo_poly.get_geometry(), obstacles=None)

    # Define the coordinates for the hole (inner polygon)

    hole = Polygon(
        [
            (12.620400, 55.688),
            (12.631, 55.689),
            (12.632, 55.687),
        ]
    )

    hole = Geometries.GeoPolygon(hole)
    hole.set_crs("EPSG:3857")

    polygon_list = Geometries.decompose_polygon(
        geo_poly.get_geometry(), obstacles=hole.get_geometry()
    )
    assert polygon_list is not None


if __name__ == "__main__":
    pytest.main(["-v", "-x", "tests/test_geometries.py"])



import math

import geojson
import pyproj
import shapely
import shapely.plotting as shplt
from shapely.geometry.polygon import orient
import random
import trajgenpy.bindings as bindings
from trajgenpy import Logging

log = Logging.get_logger()


class GeoData:
    def __init__(self, geometry, crs="WGS84"):
        self.geometry = geometry
        self.crs = crs

    def set_crs(self, crs):
        if not isinstance(crs, str):
            msg = "New CRS must be a string."
            raise ValueError(msg)

        if crs != self.crs:
            # Apply the transformer to the geometry
            self._convert_to_crs(crs)
        self.crs = crs
        return self

    def _convert_to_crs(self, crs):  # noqa: ARG002
        msg = "_convert_to_crs(crs) sould be implemented in the data classes!"
        raise NotImplementedError(msg)

    def is_geometry_of_type(self, geometry, expected_class):
        if expected_class and not isinstance(geometry, expected_class):
            msg = f"Geometry must be a {expected_class.__name__}."
            raise ValueError(msg)

    def get_geometry(self):
        return self.geometry

    def buffer(self, distance, quad_segs=1, cap_style="square", join_style="bevel"):
        self.geometry = self.geometry.buffer(distance, quad_segs, cap_style, join_style)
        return self

    def __str__(self):
        return f"Geometry in CRS: {self.crs}\nGeometry: {self.geometry}"

    def __geo_interface__(self):
        return self.geometry.__geo_interface__

    def to_geojson(self, id=None, name=None, properties=None):
        if id is None:
            id = random.randint(0, 1000000000)
        if name is None:
            name = str(id)
        if properties is None:
            properties = {}

        properties["crs"] = self.crs
        properties["name"] = name
        return geojson.Feature(id, self.geometry, properties=properties)


class GeoTrajectory(GeoData):
    def __init__(self, geometry, crs="WGS84"):
        self.is_geometry_of_type(geometry, shapely.LineString)
        super().__init__(geometry, crs)

    def plot(self, ax=None, add_points=True, color=None, linewidth=2, **kwargs):
        if self.crs == "WGS84":
            log.warning(
                "Plotting in WGS84 is not recomended as this distorts the geometry!"
            )
        shplt.plot_line(self.geometry, ax, add_points, color, linewidth, **kwargs)

    def _convert_to_crs(self, crs):
        transformer = pyproj.Transformer.from_crs(self.crs, crs, always_xy=True)

        converted_coords = [
            transformer.transform(x, y) for x, y in list(self.geometry.coords)
        ]
        self.geometry = shapely.LineString(converted_coords)


class GeoMultiTrajectory(GeoData):
    def __init__(
        self,
        geometry: (
            shapely.MultiLineString
            | list[shapely.LineString]
            | list[GeoTrajectory]
            | shapely.LineString
        ),
        crs="WGS84",
    ):
        super().__init__(geometry, crs)
        if isinstance(geometry, list):
            for line in geometry:
                self.is_geometry_of_type(line, shapely.LineString)

            super().__init__(shapely.MultiLineString(geometry), crs)
        elif isinstance(geometry, shapely.LineString):
            self.is_geometry_of_type(geometry, shapely.LineString)
            super().__init__(shapely.MultiLineString([geometry]), crs)
        elif isinstance(geometry, GeoTrajectory):
            self.is_geometry_of_type(geometry, GeoTrajectory)
            super().__init__(shapely.MultiLineString([geometry.geometry]), crs)
        else:
            self.is_geometry_of_type(geometry, shapely.MultiLineString)
            super().__init__(geometry, crs)

    def plot(self, ax=None, add_points=False, color=None, linewidth=2, **kwargs):
        if self.crs == "WGS84":
            log.warning(
                "Plotting in WGS84 is not recomended as this distorts the geometry!"
            )
        for line in self.geometry.geoms:
            shplt.plot_line(line, ax, add_points, color, linewidth, **kwargs)

    def _convert_to_crs(self, crs):
        transformer = pyproj.Transformer.from_crs(self.crs, crs, always_xy=True)
        # Convert the coordinates of each line in the MultiLineString
        converted_geoms = [
            [
                transformer.transform(x, y) for x, y in list(line.coords)
            ]  # Convert the coordinates of each line in the MultiLineString
            for line in self.geometry.geoms
        ]
        self.geometry = shapely.MultiLineString(converted_geoms)


class GeoPoint(GeoData):
    def __init__(self, geometry: shapely.Point, crs="WGS84"):
        self.is_geometry_of_type(geometry, shapely.Point)
        super().__init__(geometry, crs)

    def plot(self, ax=None, add_points=True, color=None, linewidth=2, **kwargs):
        if self.crs == "WGS84":
            log.warning(
                "Plotting in WGS84 is not recomended as this distorts the geometry!"
            )
        shplt.plot_points(self.geometry, ax, add_points, color, linewidth, **kwargs)

    def _convert_to_crs(self, crs):
        transformer = pyproj.Transformer.from_crs(self.crs, crs, always_xy=True)
        x, y = transformer.transform(self.geometry.x, self.geometry.y)
        self.geometry = shapely.Point(x, y)


class GeoPolygon(GeoData):
    def __init__(self, geometry: shapely.Polygon | shapely.LineString, crs="WGS84"):
        self.is_geometry_of_type(geometry, shapely.Polygon | shapely.LineString)
        geometry = shapely.Polygon(geometry)
        super().__init__(geometry, crs)

    def _convert_to_crs(self, crs):
        transformer = pyproj.Transformer.from_crs(self.crs, crs, always_xy=True)
        # Convert each point in the polygon
        exterior = [
            transformer.transform(x, y) for x, y in self.geometry.exterior.coords
        ]
        interiors = [
            [transformer.transform(x, y) for x, y in interior.coords]
            for interior in self.geometry.interiors
        ]
        self.geometry = shapely.Polygon(exterior, interiors)

    def plot(
        self,
        ax=None,
        add_points=False,
        color=None,
        facecolor=None,
        edgecolor=None,
        linewidth=2,
        **kwargs,
    ):
        if self.crs == "WGS84":
            log.warning(
                "Plotting in WGS84 is not recomended as this distorts the geometry!"
            )
        shplt.plot_polygon(
            polygon=self.geometry,
            ax=ax,
            add_points=add_points,
            color=color,
            facecolor=facecolor,
            edgecolor=edgecolor,
            linewidth=linewidth,
            **kwargs,
        )


class GeoMultiPolygon(GeoData):
    def __init__(self, geometry, crs="WGS84"):
        if isinstance(geometry, list):
            for geom in geometry:
                self.is_geometry_of_type(geom, shapely.Polygon)
            geometry = shapely.MultiPolygon(geometry)
        else:
            self.is_geometry_of_type(geometry, shapely.MultiPolygon)
        super().__init__(geometry, crs)

    def _convert_to_crs(self, crs):
        transformer = pyproj.Transformer.from_crs(self.crs, crs, always_xy=True)
        polygon_list = []
        for polygon in list(self.geometry.geoms):
            # Convert each point in the polygon
            exterior = [transformer.transform(x, y) for x, y in polygon.exterior.coords]
            interiors = [
                [transformer.transform(x, y) for x, y in interior.coords]
                for interior in polygon.interiors
            ]
            polygon_list.append(shapely.Polygon(exterior, interiors))

        self.geometry = shapely.MultiPolygon(polygon_list)

    def plot(
        self,
        ax=None,
        add_points=False,
        color=None,
        facecolor=None,
        edgecolor=None,
        linewidth=2,
        **kwargs,
    ):
        if self.crs == "WGS84":
            log.warning(
                "Plotting in WGS84 is not recomended as this distorts the geometry!"
            )
        shplt.plot_polygon(
            polygon=self.geometry,
            ax=ax,
            add_points=add_points,
            color=color,
            facecolor=facecolor,
            edgecolor=edgecolor,
            linewidth=linewidth,
            **kwargs,
        )


def multi_polygon_to_polygon_with_holes(multi_polygon):
    # Create an empty CGAL PolygonWithHoles
    polygon_with_holes = bindings.PolygonWithHoles(bindings.Polygon_2())

    # Iterate through each polygon in the MultiPolygon
    for polygon in multi_polygon:
        # Convert the Shapely polygon to a list of CGAL points
        cgal_points = [
            bindings.Point_2(point.x, point.y) for point in polygon.exterior.coords
        ]

        # Create a CGAL Polygon_2 from the list of points
        cgal_polygon = bindings.Polygon_2(cgal_points)

        # Add the polygon to the PolygonWithHoles as a hole
        polygon_with_holes.add_hole(cgal_polygon)

    # Return the resulting PolygonWithHoles
    return polygon_with_holes


def is_convex(polygon: shapely.Polygon):
    coords = polygon.exterior.coords
    num_coords = len(coords)

    if num_coords < 4:
        # A polygon with less than 4 vertices cannot be convex
        return False

    # Calculate the orientation of the first three points
    orientation = 0
    for i in range(num_coords):
        x1, y1 = coords[i]
        x2, y2 = coords[(i + 1) % num_coords]
        x3, y3 = coords[(i + 2) % num_coords]

        # Calculate the cross product of the vectors (x2-x1, y2-y1) and (x3-x2, y3-y2)
        cross_product = (x2 - x1) * (y3 - y2) - (y2 - y1) * (x3 - x2)

        if cross_product != 0:
            orientation = cross_product
            break

    # Check the orientation of the remaining vertices
    for i in range(num_coords):
        x1, y1 = coords[i]
        x2, y2 = coords[(i + 1) % num_coords]
        x3, y3 = coords[(i + 2) % num_coords]

        cross_product = (x2 - x1) * (y3 - y2) - (y2 - y1) * (x3 - x2)

        if cross_product * orientation < 0:
            return False

    return True


def shapely_polygon_to_cgal(polygon: shapely.Polygon):
    # Exctract all the points except the last, as this is the same as the first
    cgal_points = [
        bindings.Point_2(point[0], point[1]) for point in polygon.exterior.coords[:-1]
    ]
    # Create a CGAL Polygon_2 from the list of points
    return bindings.Polygon_2(cgal_points)


def get_sweep_offset(overlap=0.1, height=10, field_of_view=90):
    if overlap < 0 or overlap > 1:
        msg = "Overlap percentage has to be a float between 0 and 1!"
        raise ValueError(msg)

    return abs(
        2 * height * math.tan((field_of_view * math.pi / 180.0) / 2) * (1 - overlap)
    )


def generate_sweep_pattern(
    polygon: shapely.Polygon,
    sweep_offset,
    clockwise=True,
    connect_sweeps=False,
):
    # Make sure that the orientation of the polygon is counterclockwise and the interior is clockwise
    cgal_poly = shapely_polygon_to_cgal(orient(polygon=polygon))
    segments = bindings.generate_sweeps(
        cgal_poly, sweep_offset, clockwise, connect_sweeps
    )

    if connect_sweeps:
        # Combine all segments into a single LineString
        combined_line = []
        for seg in segments:
            combined_line.append([seg.source.x, seg.source.y])
            combined_line.append([seg.target.x, seg.target.y])
        result = [shapely.LineString(combined_line)]
    else:
        lines = [
            shapely.LineString(
                [[seg.source.x, seg.source.y], [seg.target.x, seg.target.y]]
            )
            for seg in segments
        ]
        result = lines

    return result


def decompose_polygon(
    boundary: shapely.Polygon, obstacles: shapely.MultiPolygon | shapely.Polygon = None
):
    if obstacles is not None:
        if isinstance(obstacles, shapely.Polygon):
            obstacles = shapely.MultiPolygon([obstacles])
        elif not isinstance(obstacles, shapely.MultiPolygon):
            msg = "Obstacles must be a Shapely MultiPolygon."
            raise ValueError(msg)

        # If the obstacles intersect with the boundary, take the union of the two and remove it from the obstacles list
        updated_obstacles = []
        for obstacle in obstacles.geoms:
            if obstacle.intersects(boundary.boundary):
                log.debug(
                    "Obstacles intersect with the boundary, the geometries will be merged."
                )
                boundary = obstacles.union(boundary)
            else:
                updated_obstacles.append(obstacle)

        obstacles = shapely.MultiPolygon(updated_obstacles)
    pwh = bindings.Polygon_with_holes_2(shapely_polygon_to_cgal(boundary))
    if obstacles is not None:
        for poly in obstacles.geoms:
            pwh.add_hole(shapely_polygon_to_cgal(poly))
    decompose_polygons = bindings.decompose(pwh)
    return [
        shapely.Polygon([(vertex.x, vertex.y) for vertex in polygon])
        for polygon in decompose_polygons
    ]



# TODO create a functional programming interface for generating tasks. This includes
#   - Polygon sweeps, both for convex and non convex polygons
#   - Task generator for different categories: Roads, boundaries, buildings
#   - Environment generator, which can define a search area based on a polygon incasing the area.
#     The environment should take potental dangerous elements in the terrain into account:
#       * Powerlines, tall buildings, etc. (This should be configurable)


def generate_search_tasks(geometries):
    # normalize_coordinates(geometries)
    pass



import geojson
import matplotlib.pyplot as plt
import shapely
import trajgenpy.Logging
from trajgenpy import Utils
from trajgenpy.Geometries import GeoMultiPolygon, GeoMultiTrajectory, GeoPolygon
from trajgenpy.Query import query_features

log = trajgenpy.Logging.get_logger()

if __name__ == "__main__":
    # Download the water features data within the bounding box
    tags = {
        "natural": ["water", "wetland"],
        # "natural": ["coastline"],
        "landuse": ["farmland"],
        "highway": [
            "primary",
            "secondary",
            "secondary_link",
            "tertiary",
            "unclassified",
            "road",
            "service",
            "path",
            "track",
        ],
        "building": ["industrial", "yes", "storage_tank"],
    }

    # Amagerværket: shapely.Polygon([(12.620400,55.687962),(12.632788,55.691589),(12.637446,55.687689),(12.624924,55.683489)])
    # Davinde: shapely.Polygon([(10.490913, 55.315346), (10.576744, 55.315346), (10.576744, 55.337417), (10.490913, 55.337417)])
    polygon = GeoPolygon(
        shapely.Polygon(
            [
                (10.490913, 55.315346),
                (10.576744, 55.315346),
                (10.576744, 55.337417),
                (10.490913, 55.337417),
            ]
        ),
        crs="WGS84",
    )  # Download the water features data within the bounding box

    features = query_features(polygon, tags)
    polygon.set_crs("EPSG:2197")

    polygon.plot(facecolor="none", edgecolor="black", linewidth=2)

    # Plot highways
    roads = GeoMultiTrajectory(features["highway"], crs="WGS84").set_crs("EPSG:2197")
    roads.plot()

    # Plot Buildings
    buildings = GeoMultiPolygon(features["building"], crs="WGS84").set_crs("EPSG:2197")

    buildings.plot(color="red")

    # Plot natural features
    # coastline = GeoMultiTrajectory(features["natural"], crs="WGS84").set_crs(
    #     "EPSG:2197"
    # )
    # coastline.plot(color="green")

    # Export the polygon and the buildings as a GeoJSON file
    geojson_collection = geojson.FeatureCollection(
        [
            polygon.to_geojson(id="boundary"),
            buildings.to_geojson(id="obstacles"),
        ]
    )

    with open("environment.geojson", "w") as f:
        geojson.dump(geojson_collection, f)

    # Plot on a map
    Utils.plot_basemap(crs="EPSG:2197")

    # No axis on the plot
    plt.axis("off")
    plt.show()



import inspect
import logging
from pathlib import Path

import colorama

LOGGER_NAME = "trajgenpy"


def init_logger():
    # Initialize colorama to support ANSI color codes on Windows
    colorama.init()

    # Define custom log levels and their corresponding colors
    LOG_COLORS = {
        "TRACE": colorama.Fore.MAGENTA,
        "DEBUG": colorama.Fore.BLUE,
        "INFO": colorama.Fore.GREEN,
        "WARNING": colorama.Fore.YELLOW,
        "ERROR": colorama.Fore.RED,
    }

    class ColoredFormatter(logging.Formatter):
        def format(self, record):
            log_level = record.levelname
            color = LOG_COLORS.get(log_level, "")
            reset = colorama.Style.RESET_ALL
            frame_info = inspect.stack()[-1]
            filename = Path(frame_info[1])
            record.filename = filename.name
            record.levelname = f"{color}{log_level}{reset}"
            return super().format(record)

    logger = logging.getLogger(LOGGER_NAME)
    log_format = "%(asctime)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    formatter = ColoredFormatter(log_format)
    logger.setLevel(logging.DEBUG)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    logger.addHandler(stream_handler)
    return logger


def get_logger():
    return logging.getLogger(LOGGER_NAME)


if get_logger().handlers == []:
    init_logger()


# def debug(*args, **kwargs):
#     message = " ".join(map(str, args))
#     logging.getLogger(LOGGER_NAME).debug(message, **kwargs)


# def info(*args, **kwargs):
#     message = " ".join(map(str, args))
#     logging.getLogger(LOGGER_NAME).info(message, **kwargs)


# def warning(*args, **kwargs):
#     message = " ".join(map(str, args))
#     logging.getLogger(LOGGER_NAME).warning(message, **kwargs)


# def error(*args, **kwargs):
#     message = " ".join(map(str, args))
#     logging.getLogger(LOGGER_NAME).error(message, **kwargs)


# if __name__ == "__main__":
#     debug("This is a debug message")
#     info("This is an info message")
#     warning("This is a warning message")
#     error("This is an error message")



import contextily as ctx
import matplotlib.pyplot as plt
import shapely
from shapely.affinity import translate

# Example usage:


# A function to plot a map using the contextily library
def plot_basemap(ax=None, provider=ctx.providers.Esri.WorldImagery, crs="WGS84"):
    if ax is None:
        ax = plt.gca()
    return ctx.add_basemap(ax, source=provider, crs=crs)


def normalize_coordinates(boundary, geometries=None):
    normalized_geoms = []
    translation_vector = shapely.Point(boundary.bounds[0], boundary.bounds[1])
    for geom in list(geometries.geoms):
        normalized_geoms.append(
            translate(geom, xoff=-translation_vector.x, yoff=-translation_vector.y)
        )

    return normalized_geoms


# lines = MultiLineString(
#     [LineString([(100, 100), (200, 200)]), LineString([(300, 300), (400, 400)])]
# )
# image_width = 800
# image_height = 600
# resulting_image = create_image_from_multilinestring(lines, image_width, image_height)


# import cv2
# def create_image_from_multilinestring(
#     multilinestring: shapely.MultiLineString,
#     thickness=5,
#     line_sigma=5,
#     intensity=10,
#     line_color=(0, 0, 255),
# ):
#     # Create a blank white image
#     image_width = int(multilinestring.bounds[2]) + 100
#     image_height = int(multilinestring.bounds[3]) + 100
#     image = np.zeros((image_height, image_width), np.uint8)

#     # Convert MultiLineString to a list of LineStrings
#     linestrings = list(multilinestring.geoms)
#     for linestring in linestrings:
#         coords = np.array(linestring.coords)

#         # Iterate over each pair of points and interpolate
#         for i in range(len(coords) - 1):
#             x1, y1 = coords[i]
#             x2, y2 = coords[i + 1]

#             # Create a Gaussian line kernel
#             line_kernel = np.zeros((image_height, image_width), np.uint8)
#             # cv2.rectangle(
#             #     line_kernel,
#             #     (int(x1), int(y1)),
#             #     (int(x2), int(y2)),
#             #     intensity,
#             #     thickness,
#             # )

#             cv2.line(
#                 line_kernel,
#                 (int(x1), int(y1)),
#                 (int(x2), int(y2)),
#                 intensity,
#                 thickness,
#             )

#             # Apply Gaussian blur to the line kernel
#             line_kernel = gaussian_filter(line_kernel, sigma=line_sigma)
#             # Add the colored line to the image
#             image = np.clip(
#                 image.astype(np.int64) + line_kernel.astype(np.int64), 0, 255
#             ).astype(np.uint8)

#     cv2.flip(image, 0, image)
#     image = cv2.normalize(image, image, 0, 255, cv2.NORM_MINMAX)

#     image = cv2.applyColorMap(image, cv2.COLORMAP_JET)
#     # Display and save the resulting image
#     # cv2.imshow("Image with Colored Gaussian Lines", image)
#     # cv2.waitKey(0)
#     # cv2.destroyAllWindows()
#     cv2.imwrite("image_with_colored_gaussian_lines.png", image)

#     return image


# # trajectory_list = Utils.normalize_coordinates(
# #     geo_poly.geometry, multi_traj.get_geometry()
# # )

# # Utils.create_image_from_multilinestring(
# #     shapely.MultiLineString(trajectory_list),
# #     thickness=int(offset * 1.2),
# #     intensity=30,
# # )



import pytest
import trajgenpy.bindings as bindings


def test_create_polygon():
    # Construct a polygon from a list of points
    points = [
        bindings.Point_2(0, 0),
        bindings.Point_2(1, 0),
        bindings.Point_2(1, 1),
        bindings.Point_2(0, 1),
    ]
    polygon = bindings.Polygon_2(points)

    # Check if the polygon is simple and convex
    assert polygon.is_simple()  # True
    assert polygon.is_convex()  # True


def test_create_sweeps():
    # create the outer polygon with holes
    outer_boundary = bindings.Polygon_2()
    outer_boundary.push_back(bindings.Point_2(0, 0))
    outer_boundary.push_back(bindings.Point_2(0, 10))
    outer_boundary.push_back(bindings.Point_2(10, 10))
    outer_boundary.push_back(bindings.Point_2(10, 0))

    outer_poly = bindings.Polygon_with_holes_2(outer_boundary)

    inner_boundary = bindings.Polygon_2()
    inner_boundary.push_back(bindings.Point_2(2, 2))
    inner_boundary.push_back(bindings.Point_2(2, 8))
    inner_boundary.push_back(bindings.Point_2(8, 8))
    inner_boundary.push_back(bindings.Point_2(8, 2))

    outer_poly.add_hole(inner_boundary)

    # create the list of polygons
    poly_list = []
    polygon1 = bindings.Polygon_2()
    polygon1.push_back(bindings.Point_2(3, 3))
    polygon1.push_back(bindings.Point_2(3, 5))
    polygon1.push_back(bindings.Point_2(5, 5))
    polygon1.push_back(bindings.Point_2(5, 3))
    poly_list.append(polygon1)

    polygon2 = bindings.Polygon_2()
    polygon2.push_back(bindings.Point_2(6, 6))
    polygon2.push_back(bindings.Point_2(6, 8))
    polygon2.push_back(bindings.Point_2(8, 8))
    polygon2.push_back(bindings.Point_2(8, 6))
    poly_list.append(polygon2)

    decomposed_polygons = bindings.decompose(outer_poly)
    segments = []
    for poly in decomposed_polygons:
        if poly.is_convex():
            segments.extend(bindings.generate_sweeps(poly, 0.5))
        else:
            msg = "not able to generate a plan for a non convex polygon"
            raise Exception(msg)
    assert len(segments) == 20


# Run the tests
if __name__ == "__main__":
    pytest.main(["-v", "-x", "tests/test_bindings.py"])



from pathlib import Path

import osmnx as ox
import shapely
from geojson import Feature, FeatureCollection, dump

from trajgenpy import Logging
from trajgenpy.Geometries import GeoPolygon

log = Logging.get_logger()


def query_features(area: GeoPolygon, tags: dict):
    # Check that the geometry has the right crs
    if not area.crs and area.crs != "WGS84":
        msg = "The geometry must use WGS84 CRS!"
        raise ValueError(msg)

    try:
        geometries = ox.features_from_polygon(area.get_geometry(), tags=tags)
    except Exception as e:
        log.error("Something went wrong while trying to query from OSM: ", extra=str(e))
        return []
    results = {tag: [] for tag in tags}
    # Iterate through the features and populate the results dictionary
    for tag in tags:
        # Filter the GeoDataFrame by the specified tag
        filtered_features = geometries[geometries[tag].notna()]
        results[tag] = shapely.intersection(
            area.get_geometry(), filtered_features.geometry.unary_union
        )
        # TODO Also include the metadata of the features for future analysis

    log.info("Extracted %d features with the tags: %s", len(geometries), str(tags))
    return results


def export_as_geojson(boundary: shapely.Polygon, obstacles, features, crs):
    boundary_feature = Feature(geometry=boundary, id="boundary")
    obstacles_feature = Feature(
        geometry=shapely.MultiPolygon(obstacles), id="obstacles"
    )
    task_feature = Feature(geometry=shapely.MultiLineString(features), id="tasks")
    feature_collection = FeatureCollection(
        [boundary_feature, task_feature, obstacles_feature], crs=crs
    )

    # Write the FeatureCollection to a GeoJSON file
    with Path("output.json").open("w") as geojson_file:
        dump(feature_collection, geojson_file)
